//  Copyright (C) 2001-2003, Evolution Robotics, Inc.
//  Any reproduction is strictly prohibited without the explicit
//  written permission of Evolution Robotics, Inc.  All rights reserved.


/**
 * @file    evo-driver/tests/diff_drive.hpp
 * @author  Copyright (C) 2002, Evolution Robotics, Inc.
 *
 * Unit tests of Diff2Drive. Currently these tests are quite dependent
 * on ResourceManager, because the drive system is an aggregate
 * resource that must search for its component resources. But to avoid
 * this dependency, a large subset of ResourceManager's functionality
 * would have to be duplicated in some test IResourceContainer. It's
 * not clear that the effort is at all worthwhile.
 */

// includes for evolution motors/IRs
#include <cmath>
#include <signal.h>
#include <evolution/Resource.hpp>
#include <evolution/Util.hpp>

// includes for server
#define WIN32_LEAN_AND_MEAN
//#include <winsock2.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "range_sensor.cpp"

// globals for server
#define DEFAULT_PORT 5001
#define DEFAULT_PROTO SOCK_STREAM // TCP
//#define DEFAULT_PROTO SOCK_DGRAM //UDP

// global declarations for IR
#define SENSOR_NAME         "Range sensor"
#define SENSOR_INTERFACE    "Evolution.IRangeSensor"

using namespace std;

// Convenient typedefs for Evolution types
typedef Evolution::Result              Result;
typedef Evolution::ResourceManager     ResourceManager;
typedef Evolution::ResourceCallbackImpl ResourceCallbackImpl;
typedef Evolution::ResourceConfigParser ResourceConfigParser;
typedef Evolution::IResourceContainer  IResourceContainer;
typedef Evolution::IResource           IResource;
typedef Evolution::IDriveSystem        IDriveSystem;
typedef Evolution::CallbackId          CallbackId;
typedef Evolution::AcquirableLock      AcquirableLock;
typedef Evolution::IConditionVariable  IConditionVariable;
typedef Evolution::DeviceBusConfig     DeviceBusConfig;
typedef Evolution::DeviceList          DeviceList;
typedef Evolution::String              String;
typedef Evolution::IRangeSensor        IRangeSensor;
typedef Evolution::Timestamp           Timestamp;

// global declarations for IRs
typedef IRangeSensor Sensor;
typedef vector<Sensor*> SensorList;
typedef vector<const char*> StringList;

SensorList sensor_list;
StringList name_list;
SensorList* sensors = &sensor_list;
StringList* names = &name_list;

// global declaratins for motors
IDriveSystem* drive = NULL;
IResourceContainer* resource_container = NULL;
ResourceManager* g_manager = NULL;
auto_ptr<ResourceManager> g_manager_cleanup;

// global declarations for server
int socket_type = DEFAULT_PROTO;
SOCKET listen_socket;
struct sockaddr_in from;
static bool shutting_down = false;

void shutdown_interrupt(int signum)
{
    signal(signum, shutdown_interrupt);
    if (signum == SIGINT)
    {
        shutting_down = true;
		cout << "SIGINT detected, shutting down...";
    }
	cout << "Interrupt detected, shutting down...";
}

void exit_handler (void)
{
    if (shutting_down)
    {
        /**if (g_manager && (g_manager->is_active ()))
        {
            g_manager->deactivate ();
        }**/
    }
}

Sensor* get_sensor(const char* sensor_id, bool print_error)
{
    // Get sensor interface
    Sensor* sensor = NULL;
    if ( resource_container->obtain_interface (0, sensor_id, 
          SENSOR_INTERFACE, (void**) &sensor) != Evolution::RESULT_SUCCESS )
    {
        if (print_error)
            std::cout << "ERROR: Resource manager failed to obtain " SENSOR_NAME 
                      << " '" << sensor_id << "' (maybe it's not setup as a " 
                      << SENSOR_NAME "?)\n";
        return NULL;
    }

    if (sensor == NULL && print_error)
        std::cout << "ERROR: " SENSOR_NAME " '" << sensor_id << "' not found.\n";

    return sensor;
}

bool get_sensor_list()//SensorList* sensors, StringList* names)
{
    cout << "Accessing all " SENSOR_NAME "s in system.\n";
	
    const ResourceConfigParser* r_config = NULL;
    if (resource_container->get_configuration(0, &r_config) != Evolution::RESULT_SUCCESS)
    {
        std::cout << "ERROR: Could not load Resource Configuration parser.\n";
        return false;
    }

    const ResourceConfigParser::BusList& bus_list = r_config->get_bus_list(); 
    for (ResourceConfigParser::BusList::const_iterator bus_iter = bus_list.begin();
         bus_iter != bus_list.end(); ++bus_iter)
    {
        //DeviceBusConfig* device_list = bus_iter.get_data();
        const DeviceBusConfig* device_bus = (*bus_iter);//.devices.begin(); 
        
        //device_list iterator.
        for (DeviceList::const_iterator device_iter = device_bus->devices.begin();
             device_iter != device_bus->devices.end(); ++device_iter) 
        {
            // Get the sensor's name
            String sensor_id;
            if ((*device_iter)->get_id(&sensor_id) != Evolution::RESULT_SUCCESS)
                continue;
			cout << "  sensor_id is " << sensor_id << endl;

			/* Warning -- the following bit of code makes it seem like a lot of
			 * sensors that actually are attached to the robot are not there...
			 * this was because we were having problems with some of them.
			 */
			// only allow the three IR sensors
			if (sensor_id == "IR_tn" || sensor_id == "IR_tne" || sensor_id == "IR_tnw" );
			// for all others, we go to the next device...
			else continue;


            // Get the sensor interface
            Sensor* sensor = get_sensor(sensor_id.c_str(), false);
            if (sensor == NULL)
                continue;

            // Track sensor in list
            const char *name = sensor_id.c_str();
            char *copy = new char[strlen(name)+1];
            strcpy(copy, name);
            sensors->push_back(sensor);
            names->push_back(copy);
            std::cout << "    " << sensor_id.c_str() << "\n";
        }
    }

    std::cout << "\n";

    return true;
}

void test_sensors(Timestamp duration, char buffer[128])
{
    cout << "Reading " SENSOR_NAME "s for " << duration << " seconds.\n";
	double ir[3] = {-1.0,-1.0,-1.0};

    duration += Evolution::Platform::get_timestamp();
    while ((Evolution::Platform::get_timestamp() < duration) && (!shutting_down))
    {
		// I think the following line simply has a bug (it needs another i< )
		//for (unsigned int i = 0; i < sensors->size() && names->size(); i++)
		//cout << "sensors->size() is " << sensors->size();
		//cout << "names->size() is " << names->size();
        for (unsigned int i = 0; i < sensors->size() && i < names->size(); i++)
        {
            // Get Distance reading
            Timestamp timestamp;
            double distance = 0.0;

            Result result = (*sensors)[i]->get_distance_reading(0, &timestamp, &distance);

            // Print result
            cout << "i=" << i << "  " << (*names)[i] << ": ";
            if (result == Evolution::RESULT_NOT_FOUND)
                cout << "Nothing detected";
            else if (result != Evolution::RESULT_SUCCESS)
                wsprintf(buffer, "READ ERROR");
			else {
				ir[i] = distance;
				cout << ir[i] << endl;
                // if ir[i] is too big, e.g., greater then 1000, we'll set it to
				// a "too big" value of 1042.
				if (ir[i] > 1000.0) {
					cout << "  Setting this \"too big\" value to a sentinel of 1042." << endl;
					ir[i] = 1042.0;
				}
			}
        }
		sprintf(buffer, "%f cm, %f cm, %f cm", ir[0], ir[1], ir[2]);

        // Don't sleep if we won't have time for another loop anyway
        double remaining = (duration - Evolution::Platform::get_timestamp());
        if (remaining <= .5)
            break;

        // Slight pause between each sensor output block
        cout << "\n";
        Evolution::Platform::millisecond_sleep(500);
    }
}

void CheckResult(bool test)
{
    cout << (test ? "passed" : "failed") << "\n";
}

void Usage(char *progname) {
    fprintf(stderr,"Usage\n%s -p [protocol] -e [endpoint] -i [interface_server]\n",
        progname);
    fprintf(stderr,"Where:\n\tprotocol is one of TCP or UDP\n");
    fprintf(stderr,"\tendpoint is the port to listen on\n");
    fprintf(stderr,"\tinterface is the ipaddr (in dotted decimal notation)");
    fprintf(stderr," to bind to\n");
    fprintf(stderr,"Defaults are TCP,5001 and INADDR_ANY\n");
    WSACleanup();
    exit(1);
}

int sensor_init(int argc, char** argv) {
	//Create list of sensors to test
    //SensorList sensor_list;
    //StringList name_list;
    //sensors = &sensor_list; // Globalized for shutdown_interrupt()
    //names = &name_list;

    if (argc <= 1) {
        // Use all sensors in system
        if (!get_sensor_list())//&sensor_list, &name_list))
            return (1);
    }
    else {
        // Use requested sensors
        for (int i = 1; i < argc; i++) {
            Sensor* sensor = get_sensor(argv[i], true);
            if (sensor) {
                const char *name = argv[i];
                char *copy = new char[strlen(name)+1];
                strcpy(copy, name);
                //name_list.push_back(copy);
                //sensor_list.push_back(sensor);
				names->push_back(copy);
                sensors->push_back(sensor);
            }
        }
    }

    if (!sensors->size()) {
        std::cout << "ERROR: No valid sensors found.\n";
        return (1);
    }
    //else
        //test_sensors(&sensor_list, &name_list, 10);
}

/* Initialize Evolution motors */
int motor_init() {
	
	// Get the interface pointer.
    std::cout << "Obtaining the drive system interface\n";
    char* interface_name = "drive";
    if (resource_container->obtain_interface (0, interface_name,
                                     IDriveSystem::INTERFACE_ID,
                                     (void**) &drive) != Evolution::RESULT_SUCCESS)
    {
        std::cout << "ERROR: Resource manager failed to obtain Drived System"
                  << " '" << interface_name << "' (maybe it's not setup as a " 
                  << "Drive system?)\n";
        return (1);
    }

    if (drive == NULL) {
        std::cout << "ERROR: Drive system '" << interface_name << "' not found.\n";
        return (1);
    }

	return 0;
}

int evolution_init(int argc, char** argv, Result result, ResourceManager* manager) {


    // Exit handling
    atexit (exit_handler);

    ResourceCallbackImpl callback;

    g_manager = manager;
    g_manager_cleanup.reset (g_manager);
    if (result != Evolution::RESULT_SUCCESS)
    {
        std::cout << "ERROR: Failed to load resource manager.\n";
        return 1;
    }

    if (manager->get_resource_container (0, &resource_container) != Evolution::RESULT_SUCCESS)
    {
        std::cout << "ERROR: Failed to get resource container from resource manager.\n";
        return 1;
    }

    if (resource_container == NULL)
    {
        std::cout << "ERROR: Got NULL resource container from resource manager.\n";
        return 1;
    }

    Evolution::Platform::millisecond_sleep(2000);

	sensor_init(argc, argv);
	motor_init();

	
}



int server_init(int argc, char** argv, unsigned short port) {

	int i, retval;
	char *interface_server= NULL;
	struct sockaddr_in local;
    WSADATA wsaData;

	/* Initialize Server */
    /* Parse arguments */
    if (argc >1) {
        for(i=1;i <argc;i++) {
            if ( (argv[i][0] == '-') || (argv[i][0] == '/') ) {
                switch(tolower(argv[i][1])) {
                    case 'p':
                        if (!stricmp(argv[i+1], "TCP") )
                            socket_type = SOCK_STREAM;
                        else if (!stricmp(argv[i+1], "UDP") )
                            socket_type = SOCK_DGRAM;
                        else
                            Usage(argv[0]);
                        i++;
                        break;

                    case 'i':
                        interface_server = argv[++i];
                        break;
                    case 'e':
                        port = atoi(argv[++i]);
                        break;
                    default:
                        Usage(argv[0]);
                        break;
                }
            }
            else
                Usage(argv[0]);
        }
    }
    
    if ((retval = WSAStartup(0x202,&wsaData)) != 0) 
	{
        fprintf(stderr,"WSAStartup failed with error %d\n",retval);
        WSACleanup();
        return -1;
    }
    
    if (port == 0)
	{
        Usage(argv[0]);
    }

    local.sin_family = AF_INET;
    local.sin_addr.s_addr = (!interface_server)?INADDR_ANY:inet_addr(interface_server); 

    /* 
     * Port MUST be in Network Byte Order
     */
    local.sin_port = htons(port);

    listen_socket = socket(AF_INET, socket_type,0); // UDP socket
    
    if (listen_socket == INVALID_SOCKET)
	{
        fprintf(stderr,"socket() failed with error %d\n",WSAGetLastError());
        WSACleanup();
        return -1;
    }
    //
    // bind() associates a local address and port combination with the
    // socket just created. This is most useful when the application is a 
    // server that has a well-known port that clients know about in advance.
    //

    if (bind(listen_socket,(struct sockaddr*)&local,sizeof(local) ) == SOCKET_ERROR) 
	{
        fprintf(stderr,"bind() failed with error %d\n",WSAGetLastError());
        WSACleanup();
        return -1;
    }

    //
    // So far, everything we did was applicable to TCP as well as UDP.
    // However, there are certain steps that do not work when the server is
    // using UDP.
    //

    // We cannot listen() on a UDP socket.

    if (socket_type != SOCK_DGRAM) 
	{
        if (listen(listen_socket,5) == SOCKET_ERROR) 
		{
            fprintf(stderr,"listen() failed with error %d\n",WSAGetLastError());
            WSACleanup();
            return -1;
        }
    }
    printf("%s: 'Listening' on port %d, protocol %s\n",argv[0],port,
        (socket_type == SOCK_STREAM)?"TCP":"UDP");
}


bool info_handler(Result result, char Buffer[128], double* veloc) 
{
	// bool shutting_down is a static (file-only) global
	// Set up movement parameters. Use velocity passed in as pointer
	double accel = 15.0;
	double turn_veloc = 0.0;
	double turn_accel = 0.0;
	Timestamp duration = 1;
	/* Handle input from client */

	if (Buffer[0] == 'i' && Buffer[1] == 'r')
	{
		test_sensors(duration, Buffer);
		return shutting_down;
	}
	if (Buffer[0] == 'v')
	{
		char temp;
		float velocity = 0, turn_velocity = 0;
		sscanf(Buffer,"%c,%f,%f", &temp, &velocity, &turn_velocity);
		wsprintf(Buffer,"Forward");
		if (velocity > 100.0 | velocity < -100.0)
			velocity = 100 * (velocity/abs(velocity));
		if (turn_velocity > 1.0 | turn_velocity < -1.0)
			turn_velocity = turn_velocity/abs(turn_velocity);
		printf("%f, %f\n", velocity, turn_velocity);
		*veloc = velocity;
		accel = 100;
		turn_veloc = turn_velocity;
		turn_accel = 100;
	}
	if (Buffer[0] == 'k')
	{
		wsprintf(Buffer,"Backward");
		*veloc = -10;
		accel = -50;
		turn_veloc = 0;
		turn_accel = 0;
		
	}
	if (Buffer[0] == 'l')
	{
		wsprintf(Buffer,"Left");
		if (turn_veloc < 0) {
			cout << "Stopping: "; 
			result = drive->stop(0, 0, 0, Evolution::IMotorCommand::STOP_ABRUPT);
		}
		printf("cur velocity = %f", *veloc);
		turn_veloc = M_PI/6.0;
		turn_accel = M_PI/2.0;
	}
	if (Buffer[0] == 'j')
	{
		wsprintf(Buffer,"Right");
		if (turn_veloc > 0) {
			cout << "Stopping: "; 
			result = drive->stop(0, 0, 0, Evolution::IMotorCommand::STOP_ABRUPT);
		}
		printf("cur velocity = %f", *veloc);
		//std::cout << "Right 1 second: "; 
		turn_veloc = -M_PI/6.0;
		turn_accel = M_PI/2.0;
	}
	if (Buffer[0] == 'z')
	{
		wsprintf(Buffer,"Stopping");
		//std::cout << "Stopping "; 
		turn_veloc = 0;
		turn_accel = 0;	
		accel = *veloc;
		*veloc = 0;


		result = drive->stop(0, 0, 0, Evolution::IMotorCommand::STOP_ABRUPT);
	}

 	if (Buffer[0] == 'q')
	{
		wsprintf(Buffer,"Losing control....now!");
		shutting_down = true;
	}

	result = drive->move_and_turn(0, *veloc, accel, turn_veloc, turn_accel);
	CheckResult(result == Evolution::RESULT_SUCCESS);
	//result = drive->stop(0, 0, 0, Evolution::IMotorCommand::STOP_ABRUPT);
	//Evolution::Platform::millisecond_sleep (500);
	//CheckResult(result == Evolution::RESULT_SUCCESS);

	return shutting_down;
}

void readyToQuit()
{
	cout << "Enter something when you're ready to quit... ";
	string resp;
	cin >> resp;
	return;
}

int main(int argc, char **argv) 
{
    char Buffer[128];
    int retval, fromlen, i;
	double veloc = 0.0;
	bool rPr = false;				// rPr == "robot present" boolean

	cout << "Welcome! Do you want to use the real robot (1=yes, anything else=no) ";
	string answerFromUser;
	cin >> answerFromUser;
	if (answerFromUser == "1") 
	{
		rPr = true;
		printf("Moving: \n  \t v,<forward velocity>,<turning_velocity>\n ");
		printf("\t forward velocity = [-100,100]\n \t turning_velocity = [-1,1]\n");
		printf("IR values: \n \t ir\n");
	}
	else 
	{
		rPr = false;
	}

	// server variables
	char *interface_server= NULL;
    unsigned short port=DEFAULT_PORT;
    SOCKET msgsock;

	// set up containers for motors
	Result result = Evolution::RESULT_SUCCESS;
	ResourceManager* manager = NULL;
	
	if (rPr) 
	{
		manager = new ResourceManager (NULL, result);
	}

	// Set up logging.
    ERSP_LOG_SET_ROOT_PRIORITY(LOG_INFO);

    signal(SIGINT, shutdown_interrupt);
#if !defined(EVOLUTION_PLATFORM_WIN32)
    signal(SIGCHLD, shutdown_interrupt);
    signal(SIGPIPE, shutdown_interrupt);
#endif

	if (rPr) 
	{
		if (evolution_init(argc, argv, result, manager) == 1)
		{
			cout << "evolution_init returned 1" << endl;
			readyToQuit();
			return 1;
		}
	}

	if (server_init(argc, argv, port) == -1)  // see above for server_init
	{
		cout << "server_init returned -1" << endl;
		readyToQuit();
		return -1;
	}

	fromlen =sizeof(from);
	if (socket_type != SOCK_DGRAM) // TCP
	{
		msgsock = accept(listen_socket,(struct sockaddr*)&from, &fromlen);
		if (msgsock == INVALID_SOCKET) 
		{
			fprintf(stderr,"accept() error %d\n",WSAGetLastError());
			WSACleanup();
			// we should have a robot cleanup function here...
			readyToQuit();
			return -1;
		}
		printf("accepted connection from %s, port %d\n", 
					inet_ntoa(from.sin_addr),
					htons(from.sin_port)) ; 
    }
    else // UDP
	{
		//
        // accept() doesn't make sense on UDP, since we do not listen()
        //
		msgsock = listen_socket;
	}


	// main listening loop
	while(1) 
	{
        // In the case of SOCK_STREAM, the server can do recv() and 
        // send() on the accepted socket and then close it.

        // However, for SOCK_DGRAM (UDP), the server will do
        // recvfrom() and sendto()  in a loop.

        if (socket_type != SOCK_DGRAM) // this is the TCP block!! SOCK_DGRAM == UDP
		{
            retval = recv(msgsock,Buffer,sizeof (Buffer),0 );
			printf("retval is %d\n",retval);

			if (retval == SOCKET_ERROR) 
			{
				fprintf(stderr,"recv() failed, retval == SOCKET_ERROR: error %d\n",WSAGetLastError());
				closesocket(msgsock);
				shutting_down = true;
			}

			if (retval == 0) 
			{
				printf("Client closed connection (retval == 0)\n");
				closesocket(msgsock); // don't know about this
				shutting_down = true;
			}
		}
        else // this is the UDP block
		{
            retval = recvfrom(msgsock,Buffer,sizeof (Buffer),0,(struct sockaddr *)&from,&fromlen);
            printf("Received datagram from %s\n",inet_ntoa(from.sin_addr));
			if (retval == SOCKET_ERROR) 
			{
				fprintf(stderr,"recvfrom() failed: error %d\n",WSAGetLastError());
				closesocket(msgsock); //don't know about this
				continue;
			}

			if (retval == 0) 
			{
				printf("Client closed connection\n");
				closesocket(msgsock);
				continue;
			}
        }



        printf("Received %d bytes, data [%s] from client\n",retval,Buffer);

		// send buffer to info_handler to do stuff with it
		// if info_handler returns true, time to shut down system
		if (!rPr && !shutting_down)
		{
			if (Buffer[0] == 'q') {
				shutting_down = true;
				wsprintf(Buffer,"I quit! Send your own packets!");
			}
			else
				wsprintf(Buffer,"Hi there!");
		}
		else if (!shutting_down)
		{
			shutting_down = info_handler(result, Buffer, &veloc);
		}

		/* make this into a clean up robot function!!
		*/
		if (shutting_down && rPr)  // if we're shutting down with the real robot
			                       // we close down the resources first...
		{
				std::cout << "Deactivating resources\n";
	
				// Release allocated interfaces
				if (resource_container && drive) {
					resource_container->release_interface (0, drive);
					drive = NULL;
				}
		  	    if (resource_container && sensors) {
					for (unsigned int i = 0; i < sensors->size(); i++)
						resource_container->release_interface (0, (*sensors)[i]);
					sensors = NULL;
					// Delete string arrays
					if (names) {
						for (unsigned int i = 0; i < names->size(); i++)
							delete [] (*names)[i];
					} // end if names
				} // end of if resource_con
		} // end of if shutting_down && rPr

		if (shutting_down) 
		{
            printf("Terminating connection\n");
			closesocket(msgsock);
			readyToQuit();
			return 0;
        }

		/* server continues */
		double v, w;
		printf("Echoing same data back to client\n");
		if (socket_type != SOCK_DGRAM) 
		{
            retval = send(msgsock,Buffer,sizeof(Buffer),0);
			printf("message sent");
		}
		else 
		{
            retval = sendto(msgsock,Buffer,sizeof (Buffer),0,(struct sockaddr *)&from,fromlen);
		}

        if (retval == SOCKET_ERROR) 
		{
            fprintf(stderr,"send() failed: error %d\n",WSAGetLastError());
        }

        if (shutting_down) 
		{
            printf("Terminating connection\n");
			closesocket(msgsock);
			readyToQuit();
			return 0;
        }
        //else 
        //    printf("UDP server looping back for more requests\n");
		//cout << "Do you want to continue? ";
		//string resp;
		//cin >> resp;
		//if (resp == "n") 
		//{
		//	printf("Terminating connection\n");
		//	closesocket(msgsock);
		//	return -1;
		//}
		//else 
		//{
        //        continue;  // redundant, but shouldn't hurt
		//}
	}
	return (0);
}
